#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use FindBin;
use List::Util 'min';
use Lingua::EN::Numbers;
use Math::PlanePath::TerdragonCurve 118; # v.118 for xyxy_to_n_either()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon ':all';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
print "  % generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::TerdragonCurve->new;

$level % 2 == 0 or die;
my $denominator = 3**($level/2) * 2;

print <<"HERE";
  \\begin{scope}[scale=1/$denominator,yscale={sqrt(3)}]
HERE

$Printwrap = 2;
printwrap("\\fill foreach \\p in {");
$Printwrap = 4;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
my $x_one = 3**($level/2);
### $x_hi
### $y_hi

my $n_is_join = sub {
  my ($n) = @_;
  my ($x,$y) = $path->n_to_xy($n);
  my @n_list = $path->xy_to_n_list($x,$y);
  @n_list = grep {$_<=2*$n_hi} @n_list;
  my $below = grep {$_ < $n_hi} @n_list;
  my $above = grep {$_ > $n_hi} @n_list;
  if (! $above || ! $below) {
    return 0;
  }
  if ($n < $n_hi && $below == 2) {
    return 0;
  }
  if ($n > $n_hi && $above == 2) {
    return 0;
  }
  return 1;
};

my %seen_triangle;
my $triangle = sub {
  my ($x1,$y1, $x2,$y2, $x3,$y3) = @_;
  ### triangle: "$x1,$y1   $x2,$y2   $x3,$y3"

  ($x1,$y1, $x2,$y2, $x3,$y3) = triangle_normalize($x1,$y1, $x2,$y2, $x3,$y3);
  ### normal:   "$x1,$y1   $x2,$y2   $x3,$y3"

  #  x1,y1  ----  x2,y2                x3,y3
  #       \      /                    /      \
  #         x3,y3                x1,y1  ----  x2,y2

  return 0 if $seen_triangle{"$x1,$y1 $x2,$y2 $x3,$y3"}++;

  foreach (1 .. $level/2) {
    ($x1,$y1) = xy_rotate_minus60($x1,$y1);
    ($x2,$y2) = xy_rotate_minus60($x2,$y2);
    ($x3,$y3) = xy_rotate_minus60($x3,$y3);
  }
  ($x1,$y1, $x2,$y2, $x3,$y3) = triangle_normalize($x1,$y1, $x2,$y2, $x3,$y3);
  ### rotated: "$x1,$y1   $x2,$y2   $x3,$y3"

  printwrap("($x1,$y1),");
  return 1;
};

my $count = 0;
foreach my $n ($n_lo .. $n_hi-1) {
  next unless $path->_UNDOCUMENTED__n_segment_is_left_boundary($n, $level);

  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my $dx = $x2-$x1;
  my $dy = $y2-$y1;

  foreach my $rot (\&xy_rotate_plus60) {
    my ($mdx,$mdy) = $rot->($dx,$dy);
    my $xm = $x1 + $mdx;
    my $ym = $y1 + $mdy;
    my $n2 = $path->xyxy_to_n_either($x1,$y1, $xm,$ym) // 0;
    my $n3 = $path->xyxy_to_n_either($x2,$y2, $xm,$ym) // 0;
    if ($n2 >= 2*$n_hi) { $n2 = 0; }
    if ($n3 >= 2*$n_hi) { $n3 = 0; }
    if ($n2 && $n3 && ($n2 >= $n_hi || $n3 >= $n_hi)) {
      my $n_part = ($n < $n_hi);
      my $n2_part = ($n2 < $n_hi);
      my $n3_part = ($n3 < $n_hi);
      if ($n_part != $n2_part
          || $n2_part != $n3_part
          || $n3_part != $n_part) {
        $count += $triangle->($x1,$y1, $x2,$y2, $xm,$ym);
      }
    }
  }
}
printwrap("}");
printwrap(" { \\p +(0,0) -- +(2,0) -- +(1,-1) -- cycle };\n");
print "  \\end{scope}\n";

my $J = 2**($level-1);
$count == $J or die "oops count=$count J=$J";

print "% GP-Test  J($level) == $count   /* total triangles */\n";
print "  % end generated\n";
exit 0;
